;; Generated package description from ement.el  -*- no-byte-compile: t -*-
(define-package "ement" "0.17" "Matrix client" '((emacs "27.1") (map "2.1") (persist "0.5") (plz "0.6") (taxy "0.10") (taxy-magit-section "0.13") (svg-lib "0.2.5") (transient "0.3.7")) :commit "ca3b0e7da1626e692df14c56e105dc91480ef814" :authors '(("Adam Porter" . "adam@alphapapa.net")) :maintainer '("Adam Porter" . "adam@alphapapa.net") :keywords '("comm") :url "https://github.com/alphapapa/ement.el")
